/*  load.c

   5 December 2001
   modify from s1 sim 1997, 1998

   add print, printc  to output buffer  31 Dec 2001
   modify to read mspec					14 Feb 2005
   for Sx cpu simulator					11 July 2006
*/

#include "sx.h"

#define 	PRIVATE		static

int clock, ninst;
FILE *fp;

typedef struct {
	int ref;
	int arity;
	int lv;
} funrec;

funrec fun[200];
int funp;
int funidx[200];

int CP, DP;		// end of CS, end of loaded DS

void error(char *s){
	printf("error: %s\n",s);
	exit(0);
}
// dump n words starting from ads
void dump(int ads, int n){
  int i;
  for(i=ads; i<ads+n; i++)
    printf("%4d %d\n",i,M[i]);
}

PRIVATE void report(void){
  float cpi;
  if( ninst == 0 )
	  printf("0 instruction\n");
  else {
  	  cpi = ((float) clock)/ninst;
  	  printf("%d instructions, %d clocks, CPI %4.2f\n",
	  	ninst,clock,cpi);
  }
}

// static alloc memory of size a
PRIVATE int xalloc(int a){
	int b;
	b = DP;
	if( DP + a >= MAXDS )
		error("xalloc: out of memory");
	DP += a;
	return b;
}

extern char *cp2;

PRIVATE void popts(void){
	TS = M[SP];
	SP--;
}

void trap(int op, int arg){		/* special function */
	int n;
	switch(op) {
	case icEnd :
		report();
		runflag = 0;
		break;
	case icArray:
		TS = xalloc(TS);
		break;
	case icSys:
		switch(arg){
		case 1: 		// print ts
  	  		n = sprintf(cp2,"%d", TS);
			popts();
	  		cp2 += n;
	  		break;
		case 2 : 		// printc ts
			n = sprintf(cp2,"%c", TS);
			popts();
			cp2 += n;
	  		break;
		case 13:
			runflag = 0;
			break;
		default:
			error("unknown syscall");
		}
		break;
	default:
		error("unknown trap");
	}
}

void show(void){		/* show some registers */
  int i;
  printf("%d:%d ",PC,mpc);
//  disassem();
  printf("TS %d FP %d SP %d NX %d FF %d {%d %d %d %d} U %d Z%d S%d\n",
    TS,FP,SP,NX,FF,V[3],V[2],V[1],V[0],U,Z,S);
}

// read microprogram mpgm.txt
void readmpgm(void){
  int i,n;
  FILE *fp;
  char ibuf[1000];

  fp = fopen("mpgm2.txt","r");
  fgets(ibuf,1000,fp);
  sscanf(ibuf,"%d",&n);	  /* length of mpgm */
  for( i=0; i<n; i++ ) {
    fgets(ibuf,1000,fp);
    strncpy(ms,ibuf+4,mwidth);  // be careful ibuf+4
	scan1(i);
  }
  fclose(fp);
}

// read mspec.h to get microaddress
void readmspec(void){
	int i,n,v;
	FILE *fp;
	char w1[80],w2[80];

	fp = fopen("mspec2.h","r");
	i = 0;
	while(1){
		fscanf(fp,"%s%s%d",w1,w2,&v);
		if( eqs(w2,"MLEN") ){	// last symbol
			mlen = v;
			break;
		}else if( eqs(w2,"MAWIDTH") ){
			mawidth = v;
		}else if( eqs(w2,"MCWIDTH") ){
			mcwidth = v;
		}else if( w2[0] == 'a' ){
			strcpy(mads[i].name,w2+2);
			mads[i].val = v;
			i++;
		}
	}
	mads[i].name[0] = 0;			// end of symbol
	mwidth = mawidth + mcwidth;
//	for(n=0; n<i; n++)
//		printf("%s %d\n",mads[n].name,mads[n].val);
//	printf("ma %d mc %d mlen %d\n",mawidth,mcwidth,mlen);
}

PRIVATE int readM(int *mem, FILE *fi){
	int i, b, e, c;
	fscanf(fi,"%d%d",&b,&e);
	for(i=b; i<=e; i++){
		fscanf(fi,"%d",&c);
		mem[i] = c;
	}
	return e;
}

// change arg at M[ads]
PRIVATE void patch(int ads, int arg){
	M[ads] = (arg << 8) | (M[ads] & 255);
}
/*
// relocate CS from a..b, offset by MAXDS
PRIVATE void reloc(int start, int end){
	int i, op, arg;
	for(i=start; i<=end; i++){
		op = M[i] & 255;
		arg = M[i] >> 8;
		switch(op){
		case icLd:
		case icSt:
		case icCall: patch(i,arg+MAXDS); break;
		}
	}
}
*/

PRIVATE int min2(int a, int b){ return (a < b) ? a : b; }
PRIVATE int max2(int a, int b){ return (a > b) ? a : b; }

// change fun.k to fun.p.u.k, CS[start..end]
PRIVATE void resize(int start, int end){
	int i, op, u, p, k, n, x, a, v;
	n = 0;
	for(i=start; i<=end; i++){
		op = M[i] & 255;
		if( op == icFun ){
			x = funidx[n];
			a = fun[x].arity;
			v = fun[x].lv;
			k = v - a + 1;
//			k = v + 1;  // som16 v is local only
			u = min2(v,MAXV);
			p = max2(u-(v-a),0);
//			p = max2(u-v,0);
			patch(i, (p<<16) | (u<<8) | (k&255) );
			n++;
		}
	}
}
// sort funrec by ref, record index in funidx[]
void sortsym(void){
	int i,j,a,b;
	for(i=0; i<funp; i++)
		funidx[i] = i;
	for(i=funp-1; i>0; i--)
		for(j=0; j<i; j++){
			a = funidx[j];
			b = funidx[j+1];
			if(fun[a].ref > fun[b].ref){
				funidx[j] = b;
				funidx[j+1] = a;
			}
		}
	for(i=0; i<funp; i++)
		printf("%d\n",fun[funidx[i]].ref);
}

// load symbol table
// som16 format: name type ref arity local
void loadsym(FILE *fi){
	int i, n, d, ty, a, v;
	char sym[80];
	funp = 0;
	fscanf(fi,"%d",&n);
	for(i=0; i<n; i++){
//		fscanf(fi,"%d%s%d%d%d%d",&d,sym,&ty,&d,&a,&v);
		fscanf(fi,"%s%d%d%d%d",sym,&ty,&d,&a,&v);
//		if(ty == 3){
		if(ty == 6){				// somv16 obj fun
			fun[funp].ref = d;
			fun[funp].arity = a;
			fun[funp].lv = v;
			funp++;
		}
	}
}

// read obj from stdin into memory
void loadobj(char *fn){
	int a;
	FILE *fi;
	fi = fopen(fn,"r");
	if(fi == NULL){
	 printf("input file not found");
	 exit(0);
	}
//	fscanf(fi,"%d",&a);
//	if( a != OBJ_SOMv2 )
//		error("not a som-v2 object");
	a = readM(M,fi);	// code
	CP = a + 1;
	a = readM(M,fi);	// string constant, data
	DP = a + 1;
	printf("load program, last address %d\n",CP);
	printf("DP %d\n",DP);
	loadsym(fi);
	fclose(fi);
	sortsym();
//	reloc(MAXDS,CP-1);
	resize(1,CP-1);
}
